@@ -1,6 +1,6 @@
 package Protocol::IMAP::Client;
 {
-  $Protocol::IMAP::Client::VERSION = '0.003';
+  $Protocol::IMAP::Client::VERSION = '0.004';
 }
 use strict;
 use warnings;
@@ -12,7 +12,7 @@ Protocol::IMAP::Client - client support for the Internet Message Access Protocol
 
 =head1 VERSION
 
-version 0.003
+version 0.004
 
 =head1 SYNOPSIS
 
@@ -154,6 +154,12 @@ is reasonably full-featured.
 
 =back
 
+=cut
+
+use Protocol::IMAP::Fetch;
+use List::Util qw(min);
+use Try::Tiny;
+
 =head1 METHODS
 
 =cut
@@ -170,6 +176,33 @@ sub new {
 	return $self;
 }
 
+sub on_read {
+	my $self = shift;
+	my $buffref = shift;
+
+#	warn "on_read with " . $$buffref;
+	if(my $fetch = $self->{fetching}) {
+		return 0 if $fetch->on_read($buffref);
+#		warn "Finished the read!\n";
+		delete $self->{fetching};
+		return 1;
+	}
+
+	if($self->is_multi_line) {
+		$self->debug("Multi line: buffer has " . $$buffref);
+		$self->on_multi_line($buffref);
+		return 0;
+	}
+
+	if($$buffref =~ s/^(.*)[\r\n]+//) {
+		$self->on_single_line($1);
+		return 1;
+		$self->debug("Switched to multiline mode") if $self->is_multi_line;
+		return 1 if $self->is_multi_line;
+	}
+	return 0;
+}
+
 =head2 on_single_line
 
 Called when there's more data to process for a single-line (standard mode) response.
@@ -181,6 +214,7 @@ sub on_single_line {
 
 	$data =~ s/[\r\n]+//g;
 	$self->debug("Had [$data]");
+
 	if($self->in_state('ConnectionEstablished')) {
 		$self->check_greeting($data);
 	}
@@ -214,11 +248,12 @@ Called when we have multi-line data (fixed size in characters).
 =cut
 
 sub on_multi_line {
-	my ($self, $data) = @_;
+	my ($self, $buffref) = @_;
 
 	if($self->{multiline}->{remaining}) {
-		$self->{multiline}->{buffer} .= $data;
-		$self->{multiline}->{remaining} -= length($data);
+		my $chunk = substr $$buffref, 0, min($self->{multiline}->{remaining}, length $$buffref), '';
+		$self->{multiline}->{buffer} .= $chunk;
+		$self->{multiline}->{remaining} -= length $chunk;
 	}
 
 	if($self->{multiline}->{remaining} == 0) {
@@ -247,28 +282,36 @@ sub handle_untagged {
 
 Fetch untagged message data. Defines the multiline callback so that we build up a buffer for the data to process.
 
+Once we call this method, the pending message takes over input until it has managed
+to read the entire response.
+
 =cut
 
 sub untagged_fetch {
 	my $self = shift;
 	my ($idx, $data) = @_;
 	$self->debug("Fetch data: $data");
-	my ($len) = $data =~ /{(\d+)}/;
-	$self->debug("Length is " . (defined($len) ? $len : "not defined"));
-	return $self unless defined $len;
 
-	$self->debug("Expect to extract [$len]");
-	$self->{multiline} = {
-		remaining => $len,
-		buffer => '',
-		on_complete => $self->_capture_weakself(sub {
-			my ($self, $buffer) = @_;
-			$self->{message}->{$idx} = $buffer;
-		})
+	try {
+		my $fetch = Protocol::IMAP::Fetch->new;
+		$fetch->completion->on_done($self->{fetch_handler}[0]);
+		push @{$self->{fetch_stack}}, $fetch;
+		$self->{fetching} = $fetch if $fetch->on_read(\$data);
+	} catch {
+		warn "error: $_"
 	};
 	return $self;
 }
 
+sub untagged_list {
+	my $self = shift;
+	my ($idx, $data) = @_;
+	$self->debug("List data: $data");
+
+#	$fetch->completion->on_done($self->{fetch_handler}[0]);
+	return $self;
+}
+
 =head2 handle_numeric
 
 Deal with an untagged response with a numeric prefix.
@@ -317,7 +360,7 @@ sub on_not_authenticated {
 		return $self->starttls;
 	} else {
 		$self->debug("Attempt to log in");
-		$self->login($self->on_user, $self->on_pass);
+		$self->invoke_event(authentication_required => );
 	}
 }
 
@@ -392,6 +435,8 @@ Request capabilities from the server.
 
 sub get_capabilities {
 	my $self = shift;
+	my %args = @_;
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'CAPABILITY',
 		on_ok		=> $self->_capture_weakself(sub {
@@ -399,13 +444,16 @@ sub get_capabilities {
 			my $data = shift;
 			$self->debug("Successfully retrieved caps: $data");
 			$self->state('NotAuthenticated');
+			$f->done($data);
 		}),
 		on_bad		=> $self->_capture_weakself(sub {
 			my $self = shift;
 			my $data = shift;
 			$self->debug("Caps retrieval failed: $data");
+			$f->fail($data);
 		})
 	);
+	$f
 }
 
 =head2 next_id
@@ -454,7 +502,7 @@ sub send_command {
 	my %args = @_;
 	my $id = exists $args{id} ? $args{id} : $self->next_id;
 	if($self->{in_idle} && defined $id) {
-# If we're currently in IDLE mode, we have to finish the current command first by issuing the DONE command.
+		# If we're currently in IDLE mode, we have to finish the current command first by issuing the DONE command.
 		return $self->done(
 			on_ok	=> $self->_capture_weakself(sub {
 				my $self = shift;
@@ -471,8 +519,8 @@ sub send_command {
 	$self->push_waitlist($id, sub {
 		my ($status, $data) = @_;
 		my $method = join('_', 'on', lc $status);
-		$args{$method}->($data) if exists $args{$method};
-		$args{on_response}->("$status $data") if exists $args{on_response};
+		(delete $args{$method})->($data) if exists $args{$method};
+		(delete $args{on_response})->("$status $data") if exists $args{on_response};
 	}) if defined $id;
 	$self->debug("Sending [$data] to server");
 	if($self->{in_idle} && defined $id) {
@@ -481,7 +529,7 @@ sub send_command {
 		$self->done;
 	} else {
 		$self->debug("In idle?") if $self->{in_idle};
-		$self->write("$data\r\n");
+		$self->write("$data\x0D\x0A");
 		$self->{in_idle} = 1 if $args{command} eq 'IDLE';
 	}
 	return $id;
@@ -507,6 +555,8 @@ See also the L<authenticate> command, which does the same thing but via L<Authen
 
 sub login {
 	my ($self, $user, $pass) = @_;
+	my %args;
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'LOGIN',
 		param		=> qq{$user "$pass"},
@@ -515,14 +565,16 @@ sub login {
 			my $data = shift;
 			$self->debug("Successfully logged in: $data");
 			$self->state('Authenticated');
+			$f->done($data);
 		}),
 		on_bad		=> $self->_capture_weakself(sub {
 			my $self = shift;
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		})
 	);
-	return $self;
+	$f
 }
 
 =head2 check_status
@@ -538,7 +590,7 @@ sub check_status {
 	my ($mbox) = $data =~ /([^ ]+)/;
 	$mbox =~ s/^(['"])(.*)\1$/$2/;
 	foreach (qw(MESSAGES UNSEEN RECENT UIDNEXT)) {
-		$status{lc($_)} = $1 if $data =~ /$_ (\d+)/;
+		$status{lc($_)} = $1 if $data =~ /$_ (\d+)/i;
 	}
 	$self->{status}->{$mbox} = \%status;
 	return $self;
@@ -554,16 +606,18 @@ sub noop {
 	my $self = shift;
 	my %args = @_;
 
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'NOOP',
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Status completed");
-			$args{on_ok}->() if $args{on_ok};
+			$f->done;
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		}
 	);
 	return $self;
@@ -579,20 +633,53 @@ sub starttls {
 	my $self = shift;
 	my %args = @_;
 
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'STARTTLS',
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("STARTTLS in progress");
-			$args{on_ok}->() if $args{on_ok};
+			$f->done();
+			$self->invoke_event(starttls => );
 			$self->on_starttls if $self->can('on_starttls');
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("STARTTLS failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f
+}
+
+sub list {
+	my $self = shift;
+	my %args = @_;
+
+	my $f = $self->_future_from_args(\%args);
+	push @{$self->{list_handler}}, $args{on_list};
+	$self->send_command(
+		command		=> join(' ', 'LIST', ($args{reference} || '""'), ($args{mailbox} || '*')),
+		on_ok		=> sub {
+			my $data = shift;
+			shift @{$self->{list_handler}};
+			$f->done($data);
+		},
+		on_bad		=> sub {
+			my $data = shift;
+			$f->fail($data);
+		}
+	);
+	$f
+}
+
+sub _future_from_args {
+	my $self = shift;
+	my $args = shift;
+	my $f = shift || Future->new;
+	$f->on_done(delete $args->{on_ok}) if exists $args->{on_ok};
+	$f->on_fail(delete $args->{on_bad}) if exists $args->{on_bad};
+	return $f
 }
 
 =head2 status
@@ -606,20 +693,22 @@ sub status {
 	my %args = @_;
 
 	my $mbox = $args{mailbox} || 'INBOX';
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'STATUS',
 		param		=> "$mbox (unseen recent messages uidnext)",
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Status completed");
-			$args{on_ok}->($self->{status}->{$mbox}) if $args{on_ok};
+			$f->done($self->{status}->{$mbox});
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	return $f;
 }
 
 =head2 select
@@ -633,20 +722,51 @@ sub select : method {
 	my %args = @_;
 
 	my $mbox = $args{mailbox} || 'INBOX';
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'SELECT',
 		param		=> $mbox,
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Have selected");
-			$args{on_ok}->($self->{status}->{$mbox}) if $args{on_ok};
+			$f->done($self->{status}->{$mbox});
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f;
+}
+
+=head2 examine
+
+Like L</select>, but readonly.
+
+=cut
+
+sub examine : method {
+	my $self = shift;
+	my %args = @_;
+
+	my $mbox = $args{mailbox} || 'INBOX';
+	my $f = $self->_future_from_args(\%args);
+	$self->send_command(
+		command		=> 'EXAMINE',
+		param		=> $mbox,
+		on_ok		=> sub {
+			my $data = shift;
+			$self->debug("Have selected for readonly");
+			$f->done($self->{status}->{$mbox});
+		},
+		on_bad		=> sub {
+			my $data = shift;
+			$self->debug("Login failed: $data");
+			$f->fail($data);
+		}
+	);
+	$f;
 }
 
 =head2 fetch
@@ -659,22 +779,26 @@ sub fetch : method {
 	my $self = shift;
 	my %args = @_;
 
-	my $msg = exists $args{message} ? $args{message} : 1;
-	my $type = exists $args{type} ? $args{type} : 'ALL';
+	my $msg = exists($args{message}) ? $args{message} : 1;
+	my $type = exists($args{type}) ? $args{type} : 'ALL';
+	my $f = $self->_future_from_args(\%args);
+	push @{$self->{fetch_handler}}, $args{on_fetch};
 	$self->send_command(
 		command		=> 'FETCH',
 		param		=> "$msg $type",
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Have fetched");
-			$args{on_ok}->($self->{message}->{$msg}) if $args{on_ok};
+			shift @{$self->{fetch_handler}};
+			$f->done($data);
 		},
 		on_bad		=> sub {
 			my $data = shift;
-			$self->debug("Login failed: $data");
+			pop @{$self->{fetch_handler}};
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f
 }
 
 =head2 delete
@@ -688,20 +812,22 @@ sub delete : method {
 	my %args = @_;
 
 	my $msg = exists $args{message} ? $args{message} : 1;
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'STORE',
 		param		=> $msg . ' +FLAGS (\Deleted)',
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Have deleted");
-			$args{on_ok}->() if $args{on_ok};
+			$f->done;
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f
 }
 
 =head2 expunge
@@ -714,19 +840,21 @@ sub expunge : method {
 	my $self = shift;
 	my %args = @_;
 
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'EXPUNGE',
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Have expunged");
-			$args{on_ok}->() if $args{on_ok};
+			$f->done;
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Login failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f
 }
 
 =head2 done
@@ -739,46 +867,32 @@ sub done {
 	my $self = shift;
 	my %args = @_;
 
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'DONE',
 		id		=> undef,
 		on_ok		=> sub {
 			my $data = shift;
 			$self->debug("Done completed");
-			$args{on_ok}->() if $args{on_ok};
+			$f->done;
 		},
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("DONE command failed: $data");
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f
 }
 
 sub untagged_exists {
 	my $self = shift;
 	my $count = shift;
 	$self->debug("Exists @_");
-	$self->maybe_invoke_event('on_message_available', $count);
+	$self->invoke_event(message_available => $count);
 	return $self;
 }
 
-#sub maybe_invoke_event {
-#	my $self = shift;
-#	$self->fetch(
-#		message	=> $idx,
-#		type => 'RFC822.HEADER',
-#		# type => 'RFC822.HEADER RFC822.TEXT',
-#		on_ok => $self->_capture_weakself(sub {
-#			my $self = shift;
-#			my $msg = shift;
-#			$self->on_message($msg);
-#			$self->idle;
-#		})
-#	);
-#	return $self;
-#}
-
 =head2 idle
 
 Switch to IDLE mode. This will put the server into a state where it will continue to send untagged
@@ -791,6 +905,7 @@ sub idle {
 	my %args = @_;
 
 	$self->{start_idle_timer}->(%args) if $self->{start_idle_timer};
+	my $f = $self->_future_from_args(\%args);
 	$self->send_command(
 		command		=> 'IDLE',
 		on_ok		=> $self->_capture_weakself( sub {
@@ -799,16 +914,17 @@ sub idle {
 			$self->{idle_timer}->stop if $self->{idle_timer};
 			$self->{in_idle} = 0;
 			my $queued = $self->{idle_queue};
-			$self->write("$queued\r\n") if $queued;
-			$args{on_ok}->() if $args{on_ok};
+			$self->write("$queued\x0D\x0A") if $queued;
+			$f->done;
 		}),
 		on_bad		=> sub {
 			my $data = shift;
 			$self->debug("Idle failed: $data");
 			$self->{in_idle} = 0;
+			$f->fail($data);
 		}
 	);
-	return $self;
+	$f;
 }
 
 =head2 is_multi_line
